#!/usr/bin/python2.4
# db_backend.py - Implements DB access for checklib.py and friends
# Copyright (C) 2006  Christian Aichinger <Greek0@gmx.net>
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import exceptions, micro
import psycopg2 as dbms
import StringIO
import config

def transactional(f):
    def _transactional_wrapper(*args, **kwargs):
        try: 
            res = f(*args, **kwargs)
        except:
            conn.rollback()
            raise
        else:
            conn.commit()
        return res

    return _transactional_wrapper


conn = None
cu = None

@transactional
def _create_schema():
    print "Creating schema"
    cu.execute("""
        CREATE TABLE packages (
                             pkg_id SERIAL PRIMARY KEY,
                             pkg_name VARCHAR NOT NULL UNIQUE,
                             pkg_ver VARCHAR NOT NULL,
                             pkg_has_libinfo INT NOT NULL
                            );""")
    cu.execute("""
        CREATE TABLE libraries (
                              lib_id SERIAL PRIMARY KEY,
                              pkg_id INTEGER NOT NULL REFERENCES packages(pkg_id) DEFERRABLE INITIALLY DEFERRED,
                              lib_soname VARCHAR,
                              lib_filename VARCHAR NOT NULL,
                              lib_basename VARCHAR NOT NULL
                             );""")
    cu.execute("""
        CREATE TABLE symbols (
                             sym_id SERIAL PRIMARY KEY,
                             lib_id INTEGER NOT NULL REFERENCES libraries(lib_id) DEFERRABLE INITIALLY DEFERRED,
                             sym_name VARCHAR NOT NULL,
                             sym_undef INTEGER NOT NULL,
                             sym_ver VARCHAR,
                             sym_verfile VARCHAR
                            );""")
    cu.execute("""
        CREATE TABLE links (
                              lnk_id SERIAL PRIMARY KEY,
                              pkg_id INTEGER NOT NULL REFERENCES packages(pkg_id) DEFERRABLE INITIALLY DEFERRED,
                              lib_id INTEGER NOT NULL REFERENCES libraries(lib_id) DEFERRABLE INITIALLY DEFERRED,
                              lnk_filename VARCHAR NOT NULL,
                              lnk_basename VARCHAR NOT NULL
                             );""")

    cu.execute("""
        CREATE TABLE result (
                             res_id SERIAL PRIMARY KEY,
                             res_pkg_name VARCHAR NOT NULL UNIQUE,
                             res_pkg_ver VARCHAR NOT NULL,
                             res_result INTEGER NOT NULL,
                             res_flags INTEGER NOT NULL,
                             res_unneeded VARCHAR,
                             res_comment VARCHAR
                            );""")
    cu.execute("""
        CREATE TABLE check_queue (
                             chk_id SERIAL PRIMARY KEY,
                             chk_pkg_name VARCHAR NOT NULL
                            );""")
    cu.execute("""
        CREATE TABLE binaries (
                             bin_id SERIAL PRIMARY KEY,
                             bin_filename VARCHAR NOT NULL
                            );""")
    cu.execute("""
        CREATE TABLE bin_syms (
                             bsym_id SERIAL PRIMARY KEY,
                             bin_id INTEGER NOT NULL REFERENCES binaries(bin_id) DEFERRABLE INITIALLY DEFERRED,
                             sym_name VARCHAR NOT NULL
                            );""")
    cu.execute("""
        CREATE TABLE bin_libs (
                             blib_id SERIAL PRIMARY KEY,
                             bin_id INTEGER NOT NULL REFERENCES binaries(bin_id) DEFERRABLE INITIALLY DEFERRED,
                             lib_id INTEGER NOT NULL REFERENCES libraries(lib_id) DEFERRABLE INITIALLY DEFERRED
                            );""")
    cu.execute("""
        CREATE TABLE pkg_import (
                             pkg_id SERIAL PRIMARY KEY,
                             pkg_name VARCHAR NOT NULL UNIQUE,
                             pkg_ver VARCHAR NOT NULL
                            );""")
    cu.execute("""
        CREATE VIEW obsolete_packages AS
                             SELECT l.pkg_id FROM packages AS l, pkg_import AS p 
                             WHERE l.pkg_name = p.pkg_name AND l.pkg_ver != p.pkg_ver
                            ;""")

    cu.execute("""CREATE INDEX symbols_lib_id_index ON symbols (lib_id);""")
    cu.execute("""CREATE INDEX symbols_name_index ON symbols (sym_name);""")
    cu.execute("""CREATE INDEX symbols_name_mix_index ON symbols (sym_name,lib_id);""")
    cu.execute("""CREATE INDEX libraries_pkg_id_index ON libraries (pkg_id);""")
    cu.execute("""CREATE INDEX packages_pkg_name_index ON packages (pkg_name);""")

    cu.execute("""CREATE INDEX bin_syms_sym_name_index ON bin_syms (sym_name);""")

    cu.execute("""CREATE INDEX pkg_import_name ON pkg_import (pkg_name);""")
    cu.execute("""CREATE INDEX pkg_import_ver ON pkg_import (pkg_ver);""")



def _check_schema():
    cu.execute("SELECT tablename FROM pg_tables WHERE tableowner = 'greek0';")
    tables = [i for (i,) in cu.fetchall()]

    if ('packages' not in tables or
        'libraries' not in tables or
        'symbols' not in tables or
        'result' not in tables):
        return False;
    return True;

class cu_wrapper:
    def __init__(self, cu):
        self.__cu = cu
    def __getattr__(self, name):
        return getattr(self.__cu, name)
    def execute(self, query, *args):
        self.__cu.execute(query.replace('?', '%s'), *args)
    def last_id(self, table, row):
        self.__cu.execute("SELECT currval('%s_%s_seq')" % (table, row))
        return cu.fetchone()[0]
    def copy_from_strings(self, table, copy_list, columns):
        copy_data = StringIO.StringIO()
        copy_data.write(''.join(copy_list))
        copy_data.seek(0)
        self.__cu.copy_from(copy_data, table, columns = columns)
        del copy_data

def init_database(dbpath):
    global conn, cu

    if conn:
        return

    conn = dbms.connect(database = dbpath)
    cu = cu_wrapper(conn.cursor())

    if not _check_schema():
        _create_schema()


@transactional
def update_result(ures):
    if ures.res_unneeded is None:
        unneeded = None
    else:
        unneeded = ','.join(ures.res_unneeded)

    cu.execute("DELETE FROM check_queue WHERE chk_pkg_name = ?", (ures.res_pkg_name,))

    cu.execute("SELECT res_id FROM result WHERE res_pkg_name = ?", (ures.res_pkg_name,))
    rows = cu.fetchall()
    if len(rows):
        ures.id = id = rows[0][0]
        cu.execute("""
            UPDATE result
                SET res_pkg_name = ?, res_pkg_ver = ?, res_result = ?, res_flags = ?, res_unneeded = ? 
                WHERE res_id = ?
            """, (ures.res_pkg_name, ures.res_pkg_ver, ures.res_result, ures.res_flags, unneeded ,id))
    else:
        # pkg not yet in results, insert
        cu.execute("INSERT INTO result (res_pkg_name,res_pkg_ver,res_result,res_flags,res_unneeded)" +
                   "VALUES (?,?,?,?,?);",
                   (ures.res_pkg_name, ures.res_pkg_ver, ures.res_result, ures.res_flags, unneeded))

        ures.id = cu.last_id('result', 'res_id')

def get_results():
    cu.execute("SELECT * FROM result;")
    rows = cu.fetchall()
    if len(rows) == 0:
        return None

    results = []
    for row in rows:
        unneeded = row[5]
        row = { 'id': row[0],
                'res_pkg_name': row[1],
                'res_pkg_ver': row[2],
                'res_result': row[3],
                'res_flags': row[4],
                'res_comment': row[6],
              }
        
        if unneeded is not None:
            unneeded = unneeded.split(',')

        row['res_unneeded'] = unneeded
        results.append(micro.uResult.from_dict(row))

    return results

def str_wrapper(value):
    if value is None:
        return r'\N'
    return value

@transactional
def add_libpkg(pkginfo):
#    cu.execute("DELETE FROM packages WHERE pkg_name = ?", (pkginfo.name,))
#    cu.execute("""
#        DELETE FROM libraries WHERE pkg_id NOT IN (SELECT pkg_id FROM packages);""")
#    cu.execute("""
#        DELETE FROM symbols WHERE lib_id NOT IN (SELECT lib_id FROM libraries);""")
#    cu.execute("""
#        DELETE FROM links WHERE pkg_id NOT IN (SELECT pkg_id FROM packages);""")

    cu.execute("INSERT INTO packages (pkg_name,pkg_ver,pkg_has_libinfo) VALUES (?,?,1);",
               (pkginfo.name, pkginfo.ver))

    pkg_id = cu.last_id('packages', 'pkg_id')
    lib_mapper = {}
    for lib in pkginfo.get_libraries():
        cu.execute("""INSERT INTO
                      libraries (pkg_id,lib_soname,lib_filename,lib_basename)
                      VALUES (?,?,?,?);""",
                   (pkg_id, lib.soname, lib.pretty_filename, lib.basename))

        lib_id = cu.last_id('libraries', 'lib_id')
        lib_mapper[lib.filename] = lib_id

        copy_list = []
        for sym in lib.get_symbols():
            if not sym.have_versym:
                ver_name = ver_file = r'\N'
            else:
                ver_name = sym.ver_name
                ver_file = sym.ver_file

            for char in ' \n\t\\':
                for string in sym.name, ver_name, ver_file:
                    if char in string:
                        raise Exception("SQL injection attempt: '%s' in '%s'" % (char, string))
            
            s = '%s\t%s\t%s\t%s\t%s\n' % (lib_id, sym.name, int(sym.undef), ver_name, ver_file)
            copy_list.append(s)

        cu.copy_from_strings('symbols', copy_list, ['lib_id','sym_name','sym_undef','sym_ver','sym_verfile'])


    for lnk in pkginfo.get_links():
        if lnk.target not in lib_mapper and lnk.target in pkginfo.elf_filenames:
            continue
        try:
            target_lib = lib_mapper[lnk.target]
        except KeyError:
            print lnk.filename, lnk.target
            print lib_mapper
            raise
        cu.execute("""INSERT INTO
                      links (pkg_id, lib_id, lnk_filename, lnk_basename)
                      VALUES (?,?,?,?);""",
                   (pkg_id, target_lib, lnk.pretty_filename, lnk.basename))

    return pkg_id

def get_libpkg(name):
    cu.execute("SELECT pkg_id FROM packages WHERE pkg_name = ? AND pkg_has_libinfo != 0;", (name,))
    rows = cu.fetchall()
    if not rows:
        return None
    return rows[0][0]

@transactional
def update_packages_list(pkgfile):
    cu.execute("DELETE FROM pkg_import;")

    for pkgname, pkg_meta in pkgfile.items():
        cu.execute("INSERT INTO pkg_import (pkg_name,pkg_ver) VALUES (?,?);", (pkgname, pkg_meta['Version']))

    conn.commit()

    cu.execute("DELETE FROM bin_libs;""")
    cu.execute("DELETE FROM bin_syms;""")
    cu.execute("DELETE FROM binaries;""")

    cu.execute("""
        DELETE FROM packages WHERE pkg_id IN (
            SELECT l.pkg_id FROM packages AS l, pkg_import AS p 
            WHERE l.pkg_name = p.pkg_name AND l.pkg_ver != p.pkg_ver
        );""")

    cu.execute("""
        DELETE FROM libraries WHERE pkg_id NOT IN (SELECT pkg_id FROM packages);""")
    cu.execute("""
        DELETE FROM symbols WHERE lib_id NOT IN (SELECT lib_id FROM libraries);""")
    cu.execute("""
        DELETE FROM links WHERE pkg_id NOT IN (SELECT pkg_id FROM packages);""")

    cu.execute("""
        DELETE FROM result WHERE res_id IN (
            SELECT res_id FROM result,pkg_import
            WHERE result.res_pkg_name = pkg_import.pkg_name AND 
                  result.res_pkg_ver != pkg_import.pkg_ver);""")

    cu.execute("""
        DELETE FROM result WHERE res_pkg_name NOT IN (SELECT pkg_name FROM pkg_import);""")

    cu.execute("""
        INSERT INTO check_queue (chk_pkg_name)
            SELECT pkg_name FROM pkg_import 
                WHERE pkg_name NOT IN (SELECT chk_pkg_name FROM check_queue) AND
                (
                 pkg_name NOT IN (SELECT res_pkg_name FROM result) OR
                 pkg_id IN 
                 (
                  SELECT pkg_id FROM pkg_import,result 
                  WHERE pkg_name = res_pkg_name AND pkg_ver != res_pkg_ver
                 )
                );""")


def get_check_queue():
    cu.execute("SELECT chk_pkg_name FROM check_queue ORDER BY chk_pkg_name;")
    rows = cu.fetchall()
    for row in rows:
        yield row[0]

def all_libs_from_scope(scope):
    qmarks = ','.join(['?'] * len(scope))

    cu.execute("""SELECT pkg_id, lib_filename, lib_basename, lib_id 
                  FROM libraries WHERE pkg_id IN (%s);""" % qmarks,
                  scope)
    libs = cu.fetchall()

    cu.execute("""SELECT pkg_id, lnk_filename, lnk_basename, lib_id
                  FROM links WHERE pkg_id IN (%s);""" % qmarks,
                  scope)
    libs += cu.fetchall()

    names = ['pkg_id', 'filename', 'basename', 'lib_id']
    libs = [dict(zip(names, lib)) for lib in libs]

    return libs

def pkg_names_from_pkg_ids(scope):
    qmarks = ','.join(['?'] * len(scope))
    cu.execute("""SELECT pkg_id, pkg_name FROM packages WHERE pkg_id IN (%s);""" % qmarks, scope)
    return cu.fetchall()


def pkg_ids_from_lib_ids(lib_ids):
    qmarks = ','.join(['?'] * len(lib_ids))
    cu.execute("""SELECT pkg_id FROM libraries WHERE lib_id IN (%s);""" % qmarks, lib_ids)

    return [i for (i,) in cu.fetchall()]


@transactional
def insert_binary(bin, load_libs):
    cu.execute("DELETE FROM bin_libs;""")
    cu.execute("DELETE FROM bin_syms;""")
    cu.execute("DELETE FROM binaries;""")

    cu.execute("""INSERT INTO binaries (bin_filename) VALUES (?);""", (bin.filename,))
    bin_id = cu.last_id('binaries', 'bin_id')

    for lib_id in load_libs:
        cu.execute("""INSERT INTO bin_libs (bin_id, lib_id) VALUES (?,?);""", (bin_id, lib_id))

    sym_map = {}
    for symbol in bin.get_symbols():
        cu.execute("""INSERT INTO bin_syms (bin_id, sym_name) VALUES (?,?);""", (bin_id, symbol.name))
        symbol.id = cu.last_id('bin_syms', 'bsym_id')
        sym_map[symbol.id] = symbol

    cu.execute("ANALYZE binaries;")
    cu.execute("ANALYZE bin_libs;")
    cu.execute("ANALYZE bin_syms;")

    return bin_id, sym_map

def find_needed_libs(bin_id):
    cu.execute("""
            SELECT DISTINCT lib_id FROM (
                SELECT DISTINCT ON (sym_name) sym_name, lib_id, blib_id 
                FROM bin_syms NATURAL JOIN (symbols NATURAL JOIN bin_libs)
                WHERE sym_undef = 0
                ORDER BY sym_name, blib_id
            ) AS bar;""")
    return [i for (i,) in cu.fetchall()]

def find_unresolved_syms(bin_id):
#    cu.execute("""
#            SELECT bsym_id FROM bin_syms WHERE sym_name NOT IN (
#                SELECT sym_name FROM symbols NATURAL JOIN bin_libs WHERE sym_undef = 0
#
#            );""")
    cu.execute("""
            SELECT bsym_id FROM bin_syms NATURAL LEFT JOIN (
                SELECT * FROM symbols NATURAL JOIN bin_libs WHERE sym_undef = 0
            ) AS foo
            WHERE lib_id IS NULL;""")
    return [i for (i,) in cu.fetchall()]
