#!/usr/bin/python2.4
# do_check.py - Backend for checklib.py, doing the actual iteration over the
#               symbol list, determining unused NEEDED entries.
# Copyright (C) 2006  Christian Aichinger <Greek0@gmx.net>
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
import os
import weakref
import hacks
import exceptions
from flags import *
from checklib import db_backend

def _find_lib_by_soname(soname, libs):
    soname = hacks.mangle_soname(soname)
    result = []
    for lib in libs:
        if soname == lib['basename']:
            result.append(lib)

    if not result:
        raise exceptions.CheckError(
              "  * Could not find file for soname '%s'" % soname)

    if len(result) == 1:
        return result[0]['lib_id']

    for path in ['/usr/lib', '/lib']:
        testname = os.path.join(path, soname)
        for lib in result:
            if lib['filename'] == testname:
                return lib['lib_id']

    one_lib = hacks.select_from_multiple_libs(soname, result)
    if one_lib:
        return one_lib['lib_id']

    print soname, result
    raise NotImplementedError("NOT IMPLEMENTED, more then 1 lib with the same soname")

def _check_binary(pkgname, bin, load_libs):
    print >>sys.stderr, "Loading..."
    bin_id, sym_map = db_backend.insert_binary(bin, load_libs)

    print >>sys.stderr, "Checking..."
    needed_lib_ids = db_backend.find_needed_libs(bin_id)
    needed_lib_ids = dict([(i,None) for i in needed_lib_ids])

    lst_ok = []
    lst_fail = []
    for lib_id in load_libs:
        if lib_id in needed_lib_ids:
            lst_ok.append(lib_id)
        else:
            lst_fail.append(lib_id)
    
    internal_problems = []
    for id in db_backend.find_unresolved_syms(bin_id):
        symbol = sym_map[id]
        if symbol.binding != 'STB_WEAK' and not hacks.sym_not_found_ok(symbol):
            internal_problems.append(symbol)

    problems = None
    if internal_problems:
        problems = InternalProblems(bin, internal_problems)

    return (lst_ok, lst_fail, problems)

class InternalProblems:
    def __init__(self, bin, symbols):
        self.bin = bin
        self.symbols = symbols
    def pretty_symbols(self):
        show_symbols = [sym.name for sym in self.symbols[:15]]
        if len(show_symbols) != len(self.symbols):
            show_symbols.append('...')

        strings = []
        inner = []
        length = 0
        for sym in show_symbols:
            sym_len = len(sym)
            if length + sym_len > 130:
                strings.append(', '.join(inner))
                length = 0
                inner = []

            # Add 2 for the ', '
            length += sym_len + 2
            inner.append(sym)

        strings.append(', '.join(inner))
        return NoticeSpace.join(strings)


class Result:
    def __init__(self, binary, res):
        self.binary = binary
        self.ok = res[0]
        self.fail = res[1]
        self.problems = res[2]

def unique(seq):
    res = {}
    for item in seq:
        res[item] = None
    return res.keys()

NoticeMessage = """
  * Info: Couldn't resolve some symbols: %s 
          NEEDED = %s""".strip('\n')

NoticeSpace = """
            """

WarnMessage = """
  * Notice: Unnecessary NEEDED entry '%s',
            probably for '%s' in pkg '%s'. No extra dependencies caused.""".strip('\n')

ErrMessage = """
  * Problem: Unnecessary NEEDED entry '%s',
             probably for '%s' in pkg '%s'. This causes an EXTRA DEPENDENCY!""".strip('\n')

def show_header():
    print ""
    print "The following issues were found:"
    print "================================"

def show_name(result):
    print "\n%s:" % result.binary.filename

def do_check(bin, scope):
    all_libs = db_backend.all_libs_from_scope(scope)

    pkg_name_map = dict(db_backend.pkg_names_from_pkg_ids(scope))
    pkg_from_lib_map = dict([(lib['lib_id'], lib['pkg_id']) for lib in all_libs])
    lib_name_map = dict([(lib['lib_id'], lib['filename']) for lib in all_libs])
    lib_basename_map = dict([(lib['lib_id'], lib['basename']) for lib in all_libs])

    results = []
    scope_names = [pkg_name_map[i] for i in scope]
    for binary in bin.get_binaries():
        load_libs = []
        not_resolved = []
        for soname in binary.needed:
            try:
                lib = _find_lib_by_soname(soname, all_libs)
            except exceptions.CheckError, exception:
                not_resolved.append(str(exception))
                continue
            load_libs.append(lib)

        if not_resolved:
            raise exceptions.CheckError('\n'.join(
                ['%s:' % binary.pretty_filename] +
                not_resolved +
                ['    Resolved Deps: %s' % ', '.join(scope_names)]))

        res = _check_binary(bin.name, binary, load_libs)
        results.append(Result(binary, res))

    print >>sys.stderr, "Analyzing..."
    really_needed_libs = []
    for result in results:
        really_needed_libs.extend(result.ok)

    really_needed_libs = unique(really_needed_libs)
    really_needed_pkgs = [bin.pkg_id] + [pkg_from_lib_map[lib] for lib in really_needed_libs]

    RES = 0
    FLAGS = 0
    showed_header = False
    force_newline = False

    if [res for res in results if res.problems]:
        print "Resolved package dependencies to:", ', '.join(scope_names)

    superfluous_pkgs_str = []
    for result in results:
        showed_name = False

        if result.binary.textrel:
            FLAGS |= FLG_TEXTREL

        if result.problems:
            FLAGS |= FLG_SYM_NOT_FOUND
            force_newline = True
            if not showed_header:
                showed_header = True
                show_header()
            if not showed_name:
                showed_name = True
                show_name(result)

            print NoticeMessage % (result.problems.pretty_symbols(), 
                                   ', '.join(result.binary.needed))

        for lib_id in result.fail:
            if not showed_header:
                showed_header = True
                show_header()
            if not showed_name:
                showed_name = True
                show_name(result)

            FLAGS |= FLG_TOO_MUCH_NEEDED 

            pkg_id = pkg_from_lib_map[lib_id]
            if pkg_id in really_needed_pkgs:
                RES |= RES_WARN
                msg = WarnMessage
            else:
                superfluous_pkgs_str.append(pkg_name_map[pkg_id])
                RES |= RES_ERR
                msg = ErrMessage
            print msg % (lib_basename_map[lib_id], lib_name_map[lib_id], pkg_name_map[pkg_id])

    if RES or force_newline:
        print # Insert newline if we outputted something

    return (RES, FLAGS, unique(superfluous_pkgs_str))
