#!/usr/bin/python2.4
# packageinfo.py - Parse all elf objects in a .deb package
# Copyright (C) 2006  Christian Aichinger <Greek0@gmx.net>
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
import os
import stat
import errno

import elf.elf as elf
import elf.elfobject as elfobject
import exceptions

def walk_error(err):
    raise err

def unique(seq):
    res = {}
    for item in seq:
        res[item] = None
    return res.keys()

class LinkError(Exception):
    pass

def do_stat(filename):
    try: 
        return os.lstat(filename)
    except OSError, (errnumber, errstr):
        if errnumber == errno.ENOENT:
            return None
        else:
            raise

def strip_root(filename, root):
    if not filename.startswith(root):
        print filename, root
        assert filename.startswith(root)
    return filename[len(root):].lstrip('/')

def full_resolve(filename, root = '/'):

    components = []
    target = strip_root(filename, root)

    while True:
        target = resolve_link(os.path.join(root, target), root)
        if target is None:
            return None
        target = strip_root(target, root)

        parts = target.rsplit('/',1)
        if len(parts) == 1:
            path = ''
            file = parts[0]
        else:
            path = parts[0]
            file = parts[1]

        components.append(file)
        target = path

        if not target:
            break
    
    components.reverse()
    return os.path.join(root, *components)

def resolve_link(filename, root = '/'):
    target = filename

    while True:
        st = do_stat(target)
        if not st:
            return None

        if not stat.S_ISLNK(st[stat.ST_MODE]):
            break

        dir = os.path.dirname(target)
        target = os.readlink(target)

        if os.path.isabs(target):
            target = os.path.join(root, target.lstrip('/'))
        else:
            target = os.path.join(dir, target)

        target = os.path.normpath(target)

    return target

class Link:
    def __init__(self, dir, filename, strip_dir_len=0):
        self.dir = dir
        self.filename = filename
        self.pretty_filename = filename[strip_dir_len:]
        self.basename = os.path.basename(filename)

        self.target = full_resolve(filename, self.dir)
        if self.target is None:
            raise LinkError()

class PackageInfo:
    def __init__(self, name, ver, filename, dir):
        self.name = name
        self.ver = ver
        self.filename = filename
        self.dir = dir

        self.elf_filenames = None
        self.lnk_filenames = None

        os.mkdir(dir)
        if not os.path.exists(filename):
            print filename
            assert os.path.exists(filename)

        ret = os.spawnlp(os.P_WAIT, "dpkg-deb", "dpkg-deb", "-x", filename, dir)
        if ret != 0:
            raise exceptions.UnpackError("Could not unpack %s" % filename)

    def __del__(self):
        if self.dir:
            os.spawnlp(os.P_WAIT, "rm", "rm", "-rf", self.dir)

    def get_elf_files(self):
        if self.elf_filenames is not None:
            return

        elf_res = []
        lnk_res = []
        for dirpath, dirnames, filenames in os.walk(self.dir, True, walk_error):
            for f in filenames:
                f = os.path.join(dirpath, f)

                try:
                    is_link = stat.S_ISLNK(os.lstat(f)[stat.ST_MODE])
                except OSError, (errnumber, errstr):
                    if errnumber == errno.EACCES:
                        continue
                    raise

                if elf.elf_type(f) in [elf.ET_EXEC, elf.ET_DYN]:
                    { True: lnk_res, False: elf_res }[is_link].append(os.path.normpath(f))
        self.elf_filenames = unique(elf_res)
        self.lnk_filenames = unique(lnk_res)

    def _do_get(self, name_list, cls, modifier):
        for filename in name_list:
            try:
                obj = cls(filename, len(self.dir))
            except (elfobject.ElfError, LinkError):
                continue

            modifier(obj)
            yield obj

    def get_libraries(self):
        self.get_elf_files()
        return self._do_get(self.elf_filenames, elfobject.Library, lambda lib: lib.get_soname())
    def get_binaries(self):
        self.get_elf_files()
        return self._do_get(self.elf_filenames, elfobject.Binary, lambda bin: bin.get_dynamic())
    def get_links(self):
        self.get_elf_files()
        return self._do_get(self.lnk_filenames, 
                            lambda f,l: Link(self.dir, f, l), lambda x: None)
