/* utils.c - A more convenient layer above libelf
   This file is loosely based on readelf.c from the Red Hat elfutils
   Copyright (C) 2006 Christian Aichinger <Greek0@gmx.net>
   Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006 Red Hat, Inc.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
   */

#include <argp.h>
#include <gelf.h>
#include <inttypes.h>
#include <libelf.h>
#include <elfutils/libebl.h>
#include <libintl.h>
#include <locale.h>
#include <mcheck.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <assert.h>

#include "utils.h"
#include "output.h"

const UElf_Section UElf_Section_INIT = { 0 };

UElf_File* file_new(Elf* elf, Ebl* ebl, const char* fname) {
    UElf_File* efile = malloc(sizeof(UElf_File));

    efile->elf = elf;
    efile->ebl = ebl;
    efile->fname = fname;

    if (elf_getshnum (elf, &efile->shnum) < 0) {
        warning("could not determinate number of sections: %s", elf_errmsg(-1));
        free(efile);
        return NULL;
    }

    if (elf_getshstrndx (elf, &efile->shstrndx) < 0) {
        warning("could not determinate section header string index: %s", elf_errmsg(-1));
        free(efile);
        return NULL;
    }

    efile->scnnames = malloc(efile->shnum * sizeof(const char*));

    UElf_Section section = UElf_Section_INIT;
    while (section_iterate(efile, &section)) {
        efile->scnnames[elf_ndxscn(section.scn)] = 
            elf_strptr(elf, efile->shstrndx, section.shdr->sh_name);
    }

    return efile;
}

void file_free(UElf_File *efile) {
    free(efile->scnnames);
    free(efile);
}

void resolve_symbol(UElf_File *efile, UElf_Section *section, int ndx, UElf_Symbol *result) {
    result->index = ndx;
    result->flags = 0;

    Elf32_Word xndx;
    GElf_Sym symmem;
    GElf_Sym *sym = gelf_getsymshndx(section->data, 0, ndx, &symmem, &xndx);

    if (!sym) {
        result->flags &= ~SIF_SYMBOL;
        return;
    }

    result->flags |= SIF_SYMBOL;
    result->type = GELF_ST_TYPE(sym->st_info);
    result->bind = GELF_ST_BIND(sym->st_info);
    memcpy(&result->sym, sym, sizeof(*sym));

    if (result->type == STT_SECTION) {
        uint32_t destndx = sym->st_shndx == SHN_XINDEX ? xndx : sym->st_shndx;
        UElf_Section sec_dest = UElf_Section_INIT;
        section_from_index(efile, destndx, &sec_dest);

        if (sec_dest.shdr) {
            result->flags |= SIF_SEC_VALID;
            result->name = elf_strptr(efile->elf, efile->shstrndx, sec_dest.shdr->sh_name);
        } else {
            result->flags &= ~SIF_SEC_VALID;
            result->name = NULL;
        }
    } else {
        result->name = elf_strptr(efile->elf, section->shdr->sh_link, sym->st_name);
    }
}

bool section_find(UElf_File *efile, GElf_Word sh_type, UElf_Section* section) {

    Elf_Scn *scn = section->scn;

    while ((scn = elf_nextscn (efile->elf, scn)) != NULL) {
        GElf_Shdr *shdr = gelf_getshdr (scn, &section->smem);
  
        if (shdr == NULL)
            internal_error("Couldn't get shdr for '%s'\n", efile->fname);
            
        if (shdr->sh_type == sh_type) {
            section->scn = scn;
            section->shdr = shdr;
            return true;
        }
    }
    section->scn = NULL;
    section->shdr = NULL;
    return false;
}

bool section_from_index(UElf_File *efile, GElf_Word index, UElf_Section* section) {

    Elf_Scn *scn = elf_getscn(efile->elf, index);
    GElf_Shdr *shdr = gelf_getshdr (scn, &section->smem);

    if (shdr == NULL)
        internal_error("Couldn't get shdr for '%s'\n", efile->fname);
        
    section->scn = scn;
    section->shdr = shdr;
    return true;
}

bool section_iterate(UElf_File *efile, UElf_Section* section) {
    section->scn = elf_nextscn(efile->elf, section->scn);
    if (!section->scn)
        return false;

    section->shdr = gelf_getshdr(section->scn, &section->smem);
    if (!section->shdr) {
        internal_error("Couldn't get shdr for '%s'\n", efile->fname);
        return false;
    }

    return true;
}

void section_fill_data(UElf_File *efile, UElf_Section* section) {
    if (section->scn)
        section->data = elf_getdata(section->scn, NULL);
    else
        section->data = NULL;
}
